from sys import stderr
from typing import Type

from django.conf import settings
from django.db.models import Model
from django.utils.module_loading import autodiscover_modules
from rest_framework import serializers, viewsets
from rest_framework.routers import DefaultRouter


class RestRouter(DefaultRouter):
    _registered_serializers = {}

    def __init__(self, *args, **kwargs):
        skip_warning = kwargs.pop('skip_warning') or False
        if not skip_warning:
            print("You're instantiating RestRouter, you should probably use channels_redux.rest.router instead")
        super(RestRouter, self).__init__(*args, **kwargs)

    def get_default_base_name(self, viewset):
        """
        If `base_name` is not specified, attempt to automatically determine
        it from the viewset.
        """
        queryset = getattr(viewset, 'queryset', None)

        assert queryset is not None, '`base_name` argument not specified, and could ' \
                                     'not automatically determine the name from the viewset, as ' \
                                     'it does not have a `.queryset` attribute.'

        return queryset.model._meta.label_lower

    def register_model(self, model, serializer=None, viewset=None,
                       fields=None, exclude_fields=None, include_fields=None,
                       queryset=None, filter_func=None, url=None, permissions_classes=None):
        exclude_fields = exclude_fields or set()
        include_fields = include_fields or set()
        fields = self.get_default_fields_for_model(model, fields, exclude_fields, include_fields)

        serializer = serializer or self.get_default_serializer(model, fields)
        self._registered_serializers[model] = serializer
        if not issubclass(serializer, HyperlinkedModelSerializer):
            print("Serializer is not an instance of channels_redux.rest.HyperlinkedModelSerializer "
                  "unexpected behavior may occur", file=stderr)

        queryset = queryset or self.get_default_queryset(model)
        if viewset is None:
            viewset = self.get_default_viewset(queryset, serializer, filter_func, permissions_classes)
        else:
            viewset.serializer_class = serializer  # Ensure that the serializer on the viewset matches the one we expect

        if url is None:
            url = self.get_default_url(model)
        self.register(url, viewset)

    def get_default_queryset(self, model):
        return model.objects.all()

    def get_serializer(self, model):
        return self._registered_serializers[model]

    def get_default_url(self, model):
        return '/'.join((model._meta.app_label, model._meta.verbose_name_plural.replace(' ', '-')))

    def get_default_serializer(self, model_class, serializer_fields):
        class ModelSerializer(HyperlinkedModelSerializer):
            class Meta:
                model = model_class
                fields = serializer_fields
        return ModelSerializer

    def get_default_viewset(self, qs, serializer, filter_func, permissions):
        class ModelViewSet(viewsets.ModelViewSet):
            queryset = qs
            serializer_class = serializer
            permission_classes = permissions or tuple()

            def get_queryset(self):
                if filter_func is None:
                    return self.queryset
                return filter_func(self.queryset, self.request)
        return ModelViewSet

    def get_default_fields_for_model(self, model: Type[Model], fields, exclude_fields, include_fields):
        if fields is not None:
            updated_fields = set(fields)
            updated_fields.add('pk')
            updated_fields.add('url')
            return tuple(updated_fields)

        updated_fields = set(include_fields)
        excluded = set(exclude_fields)

        updated_fields.add('url')  # Always include this
        updated_fields |= model._meta._property_names  # This include pk as well as developer defined properties

        for field in model._meta.get_fields():
            if field.auto_created or field.name in excluded:  # Skip reverse relations
                continue
            updated_fields.add(field.name)

        return tuple(updated_fields)


def autodiscover():
    autodiscover_modules('rest', register_to=None)


router = RestRouter(skip_warning=True)


class HyperlinkedModelSerializer(serializers.HyperlinkedModelSerializer):
    @staticmethod
    def fix_view_name(field_kwargs, model):
        if 'view_name' in field_kwargs:
            view_name_suffix = field_kwargs['view_name'].split('-')[-1]
            field_kwargs['view_name'] = "{}:{}-{}".format(settings.API_APP_NAMESPACE,
                                                          model._meta.label_lower,
                                                          view_name_suffix)
        return field_kwargs

    def build_url_field(self, field_name, model_class):
        field_class, field_kwargs = super().build_url_field(field_name, model_class)
        return field_class, self.fix_view_name(field_kwargs, self.Meta.model)

    def build_relational_field(self, field_name, relation_info):
        field_class, field_kwargs = super().build_relational_field(field_name, relation_info)
        return field_class, self.fix_view_name(field_kwargs, model=relation_info[1])



import json as jsonlib

from django import template
from django.utils.safestring import mark_safe

register = template.Library()


@register.filter
def json(value):
    uncleaned = jsonlib.dumps(value)
    return mark_safe(uncleaned)


def react_template_tag(context):
    context['react_root_id'] = 'react_root'
    return context


register.inclusion_tag("react/react_root.html", takes_context=True, name="react_root")(react_template_tag)
register.inclusion_tag("react/react_scripts.html", takes_context=True, name="react_scripts")(react_template_tag)
register.inclusion_tag("react/react_full.html", takes_context=True, name="react")(react_template_tag)



import os.path
from typing import List, Union

from django.db.models import QuerySet, Model
from django.views.generic import *
from django.views.generic.base import ContextMixin

from channels_redux.rest import router


def get_app_name(obj):
    return obj.__module__.split('.')[0]


class ReactMixin(object):
    initial_react_state = {}
    react_component = None
    template_component_name = 'component'
    react_dist_directory = 'dist'
    template_state_name = 'props'
    key_field = 'url'

    def get_initial_react_state(self):
        if isinstance(self, ReactMultiModelMixin) and self.use_objects_as_state:
            return self.object_based_react_state()
        else:
            return self.initial_react_state

    def get_react_component(self):
        return os.path.join(self.react_dist_directory, get_app_name(self), self.react_component)


class ReactMultiModelMixin(object):
    use_objects_as_state = True
    querysets = None  # type: List[Union[QuerySet, Model]]

    def get_querysets(self) -> List[Union[QuerySet, Model]]:
        return self.querysets

    def object_based_react_state(self):
        return {"objects": self.get_serialized_objects()}

    def get_serialized_objects(self):
        model_object_pairs = []
        for queryset_or_object in self.get_querysets():
            if isinstance(queryset_or_object, QuerySet):
                serializer = router.get_serializer(queryset_or_object.model)
                model_name = queryset_or_object.model._meta.label_lower
                many = True
            else:
                serializer = router.get_serializer(queryset_or_object._meta.model)
                model_name = queryset_or_object._meta.label_lower
                many = False
            serialized_objects = serializer(queryset_or_object,
                                            context={'request': self.request, 'view': self, 'format': None},
                                            many=many).data
            model_object_pairs.append((model_name, serialized_objects))
        return redux_objects(self.key_field, *model_object_pairs)


class ReactSingleModelMixin(ReactMultiModelMixin):
    def get_querysets(self):
        return [self.get_queryset()]


def redux_objects(key_field, *model_object_pairs):
    object_state = {}
    for pair in model_object_pairs:
        model_name, serialized_objects = pair
        existing = object_state.get(model_name, {})
        object_state[model_name] = {
            **existing,
            **key_to_object_dict(key_field, serialized_objects)
        }
    return object_state


def key_to_object_dict(key_field, serialized_objects: Union[list, dict]):
    if isinstance(serialized_objects, list):
        return {obj[key_field]: obj for obj in serialized_objects}
    else:
        return {serialized_objects[key_field]: serialized_objects}


def get_react_context(react_view: ReactMixin):
    context = {
        react_view.template_state_name: react_view.get_initial_react_state(),
        react_view.template_component_name: react_view.get_react_component(),
        "key_field": react_view.key_field
    }
    return context


def get_full_context(react_view: ReactMixin, parent_view: ContextMixin, *args, **kwargs):
    return {
        **parent_view.get_context_data(*args, **kwargs),
        **get_react_context(react_view)
    }


class ReactView(View, ReactMixin):
    pass


class ReactRedirectView(RedirectView, ReactMixin):
    pass


class ReactTemplateView(TemplateView, ReactMixin):
    def get_context_data(self, *args, **kwargs):
        return get_full_context(self, super(), *args, **kwargs)


class ReactArchiveIndexView(ArchiveIndexView, ReactMixin, ReactSingleModelMixin):
    def get_context_data(self, *args, **kwargs):
        return get_full_context(self, super(), *args, **kwargs)


class ReactYearArchiveView(YearArchiveView, ReactMixin, ReactSingleModelMixin):
    def get_context_data(self, *args, **kwargs):
        return get_full_context(self, super(), *args, **kwargs)


class ReactMonthArchiveView(MonthArchiveView, ReactMixin, ReactSingleModelMixin):
    def get_context_data(self, *args, **kwargs):
        return get_full_context(self, super(), *args, **kwargs)


class ReactWeekArchiveView(WeekArchiveView, ReactMixin, ReactSingleModelMixin):
    def get_context_data(self, *args, **kwargs):
        return get_full_context(self, super(), *args, **kwargs)


class ReactDayArchiveView(DayArchiveView, ReactMixin, ReactSingleModelMixin):
    def get_context_data(self, *args, **kwargs):
        return get_full_context(self, super(), *args, **kwargs)


class ReactTodayArchiveView(TodayArchiveView, ReactMixin, ReactSingleModelMixin):
    def get_context_data(self, *args, **kwargs):
        return get_full_context(self, super(), *args, **kwargs)


class ReactDateDetailView(DateDetailView, ReactMixin, ReactSingleModelMixin):
    def get_context_data(self, *args, **kwargs):
        return get_full_context(self, super(), *args, **kwargs)


class ReactDetailView(DetailView, ReactMixin, ReactSingleModelMixin):
    def get_context_data(self, *args, **kwargs):
        return get_full_context(self, super(), *args, **kwargs)


class ReactFormView(FormView, ReactMixin):
    def get_context_data(self, *args, **kwargs):
        return get_full_context(self, super(), *args, **kwargs)


class ReactCreateView(CreateView, ReactMixin, ReactSingleModelMixin):
    def get_context_data(self, *args, **kwargs):
        return get_full_context(self, super(), *args, **kwargs)


class ReactUpdateView(UpdateView, ReactMixin, ReactSingleModelMixin):
    def get_context_data(self, *args, **kwargs):
        return get_full_context(self, super(), *args, **kwargs)


class ReactDeleteView(DeleteView, ReactMixin, ReactSingleModelMixin):
    def get_context_data(self, *args, **kwargs):
        return get_full_context(self, super(), *args, **kwargs)


class ReactListView(ListView, ReactMixin, ReactSingleModelMixin):
    def get_context_data(self, *args, **kwargs):
        return get_full_context(self, super(), *args, **kwargs)


class ReactMultiModelView(TemplateView, ReactMixin, ReactMultiModelMixin):
    def get_context_data(self, *args, **kwargs):
        return get_full_context(self, super(), *args, **kwargs)



from django.db.models.signals import post_save, post_delete
from django.dispatch import receiver

from channels_redux.notify import NotifyConsumer, NotifierMixin

print('-' * 80)
print('IMPORTED')
print('-' * 80)


@receiver(post_save, dispatch_uid="channels_redux.signals.object_saved")
def object_saved(sender, instance: NotifierMixin, created, **kwargs):
    if not isinstance(instance, NotifierMixin):
        return
    if created:
        NotifyConsumer.object_created(instance)
    else:
        NotifyConsumer.object_updated(instance)


@receiver(post_delete, dispatch_uid="channels_redux.signals.object_deleted")
def object_deleted(sender, instance: NotifierMixin, **kwargs):
    if not isinstance(instance, NotifierMixin):
        return
    NotifyConsumer.object_deleted(instance)



from django.apps import AppConfig


class ChannelsReduxConfig(AppConfig):
    name = 'channels_redux'

    def ready(self):
        from channels_redux import signals
        super(ChannelsReduxConfig, self).ready()
